"""
SPDX-License-Identifier: GPL-3.0-or-later
SPDX-FileCopyrightText: 2020 CERN
"""

import pytest
from PyAdcLib import *

@pytest.fixture(scope="function")
def adc100m14b4cha():
    adc = PyFmcAdc100m14b4ch(pytest.adc_id)
    yield adc
    adc.acq_stop(0)

@pytest.fixture(scope="function")
def adc100m14b4cha_def_cfg():
    adc = PyFmcAdc100m14b4ch(pytest.adc_id)
    for ch in range(4):
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_CHN, ch)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_RANGE, 0x45)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_TERMINATION, 0)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_OFFSET, 0)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_SATURATION, 0x7FFF)
        adc.apply_config(conf, 0)

    for ch in range(4):
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, ch)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_THR_ENABLE, 0)
        adc.apply_config(conf, 0)

    conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_EXT, 0)
    conf.value_set(PyAdcConf.ADC_CONF_TRG_EXT_ENABLE, 0)
    adc.apply_config(conf, 0)

    yield adc
    adc.acq_stop(0)

@pytest.fixture(scope="function")
def adc_simple():
    adc = PyFmcAdc100m14b4ch(pytest.adc_id)
    for ch in range(4):
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_CHN, ch)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_RANGE, 0x45)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_TERMINATION, 0)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_OFFSET, 0)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_SATURATION, 0x7FFF)
        adc.apply_config(conf, 0)

    for ch in range(4):
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, ch)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_THR_ENABLE, 0)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_THR_POLARITY, 0)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_THR_DELAY, 0)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_THR_THRESHOLD, 0)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_THR_HYSTERESIS, 0)
        adc.apply_config(conf, 0)

    conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_EXT, 0)
    conf.value_set(PyAdcConf.ADC_CONF_TRG_EXT_ENABLE, 0)
    adc.apply_config(conf, 0)

    conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_TIM, 0)
    conf.value_set(PyAdcConf.ADC_CONF_TRG_TIM_ENABLE, 0)
    conf.value_set(PyAdcConf.ADC_CONF_TRG_TIM_SECONDS, 0)
    conf.value_set(PyAdcConf.ADC_CONF_TRG_TIM_NANO_SECONDS, 0)
    adc.apply_config(conf, 0)

    conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_ACQ, 0)
    conf.value_set(PyAdcConf.ADC_CONF_ACQ_PRE_SAMP, 1)
    conf.value_set(PyAdcConf.ADC_CONF_ACQ_POST_SAMP, 2)
    conf.value_set(PyAdcConf.ADC_CONF_ACQ_N_SHOTS, 1)
    adc.apply_config(conf, 0)

    yield adc
    adc.acq_stop(0)

def pytest_addoption(parser):
    parser.addoption("--type", choices=[PyFmcAdc100m14b4ch.BOARD_NAME],
                     required=True, help="Fmc ADC type")
    parser.addoption("--id", type=lambda x : int(x, 16),
                     required=True, help="Fmc ADC Identifier")

def pytest_configure(config):
    pytest.adc_type = config.getoption("--type")
    pytest.adc_id = config.getoption("--id")
