/* Copyright 2013 CERN
 * Author: Federico Vaga <federico.vaga@vaga.pv.it>
 * License: GPLv2
 *
 * This is a simple program to configure the FMC ADC trigger. It is not bug
 * aware because it is only a demo program
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <errno.h>
#include <pthread.h>
#include <time.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <inttypes.h>
#include <adc-lib.h>
#include <adc-lib-100m14b4cha.h>

static int arg_show_config = 0;
static int arg_no_read = 0;
static int arg_plot = 0;
static int arg_x_display = 0;
static int arg_trgsw_delay = 0;
static int arg_trgsw = 0;
static int fixup = 0;
static int statistics = 0;

static char git_version[] = "version: " GIT_VERSION;

static const char *help_msg =
	"adc-acq -D <device-name>@0x<device-id> [OPTIONS]\n\n";
static const char *help_msg_opt =
"  --device|-D <device-name>@<device-id>: unique device identifier  (e.g.: \"fmc-adc@0x0400\")\n"
"  --acquisition|-a <parameters>        configure the acquisition\n"
"        <pre-sample>,<post-sample>,<n-shots>,<undersample>\n"
"  --trg-ext <parameters>               configure an external trigger\n"
"        <idx>,<enable>,<polarity>,<delay>\n"
"  --trg-thr <parameters>               configure a threshold trigger\n"
"        <idx>,<enable>,<polarity>,<threshold>,<hysteresis>,<delay>\n"
"  --trg-tim <parameters>               configure a timing trigger\n"
"        <idx>,<enable>,<seconds>,<nanoseconds>\n"
"  --trg-sw <parameters>                configure a software trigger\n"
"        <delay_seconds>\n"
"  --off-clr <ac-mode>,<idx>               run offset-clear on a given channel\n"
"  --channel| -c <parameters>           configure an acquisition channel\n"
"        <channel>,<termination>,<range>,<offset>,<saturation>\n"
"  --timeout|-T <millisec>  timeout for acquisition\n"
"  --binary|-B <file>       save binary to <file>\n"
"  --multi-binary|-M <file> save two files per shot: "
					"<file>.0000.ctrl etc\n"
"  --dont-read|-N           config-only\n"
"  --loop|-l <num>          number of loop before exiting\n"
"  --show-data|-s <num>     how many data to display: "
					">0 from head, <0 from tail\n"
"  --plot                   plot acquisition\n"
"  --X11|-X                 Gnuplot will use X connection\n"
"  --stats                  It prints some statistics\n"
"  --version|-V             print version information\n"
"  --help|-h                show this help\n"
"\n";

static const char *help_msg_opt_desc =
"Parameter descriptions (alphabetic order)\n"
"  <ac-mode>\n"
"    Offset auto-clear mode mode, a: automatic (use library configuration), m: manual (use user configuration)\n"
"  <enable>\n"
"    enabled: 1, disabled: 0\n"
"  <idx>\n"
"    resource index [0, N]\n"
"  <nanoseconds>\n"
"    number of nanoseconds since the card base time.\n"
"  <n-shots>\n"
"    number of consecutive acquisitions\n"
"  <offset>*\n"
"    offset to apply on input channel\n"
"  <polarity>\n"
"    positive edge/slope: 0, negative edge/slope: 1\n"
"  <post-sample>\n"
"    number of samples to be acquired after trigger (including the trigger)\n"
"  <pre-sample>\n"
"    number of samples to be acquired before trigger\n"
"  <range>*\n"
"    voltage range to use\n"
"  <saturation>*\n"
"    saturation value for an input channel\n"
"  <seconds>\n"
"    number of seconds since the card base time.\n"
"  <termination>\n"
"    enabled: 1, disabled: 0\n"
"  <threshold>*\n"
"    <adc-value> that trigger the acquisition\n"
"  <undersample>\n"
"    number of samples to skip during the acquisition\n"
"\n"
"    * The meaning depends on the board in used\n"
"\n";

static void fald_help()
{
	int i;

	fputs(help_msg, stderr);

	fprintf(stderr, " <device-name>: name of the device type to open\n");
	for (i = 0; i < __ADC_SUPPORTED_BOARDS_LAST_INDEX; i++)
		fprintf(stderr, "    %s\n", adc_board_name[i]);

	fputs(help_msg_opt, stderr);
	fputs(help_msg_opt_desc, stderr);
}


enum fald_acq_options {
	FALD_ACQ_OPT_TRG_EXT = 1024,
	FALD_ACQ_OPT_TRG_THR,
	FALD_ACQ_OPT_TRG_TIM,
	FALD_ACQ_OPT_TRG_SW,
	FALD_ACQ_OPT_OFF_CLR,
};

static struct option options[] = {
	{"device",	required_argument, 0, 'D'},
	{"acquisition", required_argument, 0, 'a'},
	{"trg-ext",	required_argument, 0, FALD_ACQ_OPT_TRG_EXT},
	{"trg-thr",	required_argument, 0, FALD_ACQ_OPT_TRG_THR},
	{"trg-tim",	required_argument, 0, FALD_ACQ_OPT_TRG_TIM},
	{"trg-sw",	required_argument, 0, FALD_ACQ_OPT_TRG_SW},
	{"channel",	required_argument, 0, 'c'},
	{"timeout",	required_argument, 0, 'T'},
	{"off-clr",	required_argument, 0, FALD_ACQ_OPT_OFF_CLR},
	{"fixup",	no_argument, &fixup, 1},
	{"stats",	no_argument, &statistics, 1},


	/* new options, to help stress-test */
	{"binary",	required_argument, 0, 'B'},
	{"multi-binary",required_argument, 0, 'M'},
	{"dont-read",	no_argument,       &arg_no_read, 1},
	{"loop",	required_argument, 0, 'l'},
	{"show-data",	required_argument, 0, 's'},
	{"plot",	no_argument, &arg_plot, 1},
	{"X11", no_argument, &arg_x_display , 1},

	/* loop for stess test */
	{"loop",	required_argument, 0, 'l'},

	{"version",	no_argument,       0, 'V'},
	{"help", 	no_argument,       0, 'h'},
	{"show-config", no_argument,       &arg_show_config, 1},
	{0, 0, 0, 0}
};

#define GETOPT_STRING "D:Vha:c:T:B:M:l:s:"

static void print_version(char *pname)
{
	printf("%s %s\n", pname, git_version);
	adc_print_version();
}

/* variables shared between threads */
static char *devname = NULL;
static unsigned int devid = 0;
static int show_ndata = INT_MAX; /* by default all values are displayed */
static int binmode;
static int timeout = -1;
static int loop = 1;
static char *basefile;
static char *_argv[16];

/* default is 1 V*/
static double bit_scale = 0.5/(1<<15);



/**
 * It shows the current configuration
 * @param[in] adc the ADC token
 */
static void fald_acq_print_config(struct adc_dev *adc)
{
	struct adc_conf cfg, cfg_brd;
	uint32_t value, n;
	int err, i;

	if (!arg_show_config)
		return;
	/*
	 * Show the configuration only the first time,
	 * not at every loop iteration
	 */
	arg_show_config = 0;

	fprintf(stdout, "Board Configuration\n");
	memset(&cfg_brd, 0, sizeof(struct adc_conf));
	cfg_brd.type = ADC_CONF_TYPE_BRD;
	adc_set_conf_mask_all(&cfg_brd, adc);
	err = adc_retrieve_config(adc, &cfg_brd);
	if (err) {
		fprintf(stderr,
			"Cannot retrieve the board configuration: (%d) %s\n",
			errno, adc_strerror(errno));
	} else {
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_STATUS, &value);
		if (!err)
			fprintf(stdout, "\tStatus: %u\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_STATE_MACHINE_STATUS, &value);
		if (!err)
			fprintf(stdout, "\tState machine: %u\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_MAX_FREQ_HZ, &value);
		if (!err)
			fprintf(stdout, "\tMax sampling frequency: %u\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_MIN_FREQ_HZ, &value);
		if (!err)
			fprintf(stdout, "\tMin sampling frequency: %u\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_CHAN, &value);
		if (!err)
			fprintf(stdout, "\tNumber of channels: %u\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_TRG_EXT, &value);
		if (!err)
			fprintf(stdout, "\tNumber of external triggers: %u\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_TRG_THR, &value);
		if (!err)
			fprintf(stdout, "\tNumber of threshold triggers: %u\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_TRG_TIM, &value);
		if (!err)
			fprintf(stdout, "\tNumber of timer triggers: %u\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_UTC_TIMING_BASE_S, &value);
		if (!err)
			fprintf(stdout, "\tTime Seconds: %u\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_UTC_TIMING_BASE_T, &value);
		if (!err)
			fprintf(stdout, "\tTime sub-second ticks: %u\n", value);
		err = adc_get_conf(&cfg_brd, ADC_CONF_UTC_TIMING_BASE_B, &value);
		if (!err)
			fprintf(stdout, "\tTime sub-second bins: %u\n", value);
	}

	fprintf(stdout, "Acquisition Configuration\n");
	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_ACQ;
	adc_set_conf_mask_all(&cfg, adc);
	err = adc_retrieve_config(adc, &cfg);
	if (err) {
		fprintf(stderr,
			"Cannot retrieve the acquisition configuration: (%d) %s\n",
			errno, adc_strerror(errno));
	} else {
		err = adc_get_conf(&cfg, ADC_CONF_ACQ_N_SHOTS, &value);
		if (!err)
			fprintf(stdout, "\tShots: %u\n", value);
		err = adc_get_conf(&cfg, ADC_CONF_ACQ_PRE_SAMP, &value);
		if (!err)
			fprintf(stdout, "\tPre samples: %u\n", value);
		err = adc_get_conf(&cfg, ADC_CONF_ACQ_POST_SAMP, &value);
		if (!err)
			fprintf(stdout, "\tPost samples: %u\n", value);
		err = adc_get_conf(&cfg, ADC_CONF_ACQ_UNDERSAMPLE, &value);
		if (!err)
			fprintf(stdout, "\tUndersample (decimation): %u\n", value);
		err = adc_get_conf(&cfg, ADC_CONF_ACQ_FREQ_HZ, &value);
		if (!err)
			fprintf(stdout, "\tSampling rate: %u Hz\n", value);
		err = adc_get_conf(&cfg, ADC_CONF_ACQ_N_BITS, &value);
		if (!err)
			fprintf(stdout, "\tBit per sample: %u\n", value);
	}

	fprintf(stdout, "Channel Configuration\n");
	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_CHN;
	adc_set_conf_mask_all(&cfg, adc);
	err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_CHAN, &n);
	if (err)
		n = 0;
	for (i = 0; i < n; ++i) {
		fprintf(stdout, "\tChannel %d\n", i + 1);
		cfg.route_to = i;
		err = adc_retrieve_config(adc, &cfg);
		if (err) {
			fprintf(stderr,
				"Cannot retrieve the external trigger configuration (): (%d) %s\n",
				errno, adc_strerror(errno));
		} else {
			err = adc_get_conf(&cfg, ADC_CONF_CHN_RANGE, &value);
			if (!err)
				fprintf(stdout, "\t\tRange: %u\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_CHN_OFFSET, &value);
			if (!err)
				fprintf(stdout, "\t\tOffset: %u\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_CHN_SATURATION, &value);
			if (!err)
				fprintf(stdout, "\t\tSaturation: %u\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_CHN_TERMINATION, &value);
			if (!err)
				fprintf(stdout, "\t\tTermination: %u\n", value);
		}
	}

	fprintf(stdout, "Trigger (External) Configuration\n");
	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_TRG_EXT;
	adc_set_conf_mask_all(&cfg, adc);
	err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_TRG_EXT, &n);
	if (err)
		n = 0;
	for (i = 0; i < n; ++i) {
		fprintf(stdout, "\tExternal trigger %d\n", i + 1);
		cfg.route_to = i;
		err = adc_retrieve_config(adc, &cfg);
		if (err) {
			fprintf(stderr,
				"Cannot retrieve the external trigger configuration (): (%d) %s\n",
				errno, adc_strerror(errno));
		} else {
			err = adc_get_conf(&cfg, ADC_CONF_TRG_EXT_ENABLE, &value);
			if (!err)
				fprintf(stdout, "\t\tEnable: %u\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_EXT_POLARITY, &value);
			if (!err)
				fprintf(stdout, "\t\tPolarity: %u\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_EXT_DELAY, &value);
			if (!err)
				fprintf(stdout, "\t\tDelay: %u\n", value);
		}
	}

	fprintf(stdout, "Trigger (Threshold) Configuration\n");
	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_TRG_THR;
	adc_set_conf_mask_all(&cfg, adc);
	err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_TRG_THR, &n);
	if (err)
		n = 0;
	for (i = 0; i < n; ++i) {
		fprintf(stdout, "\tThreshold trigger %d\n", i + 1);
		cfg.route_to = i;
		err = adc_retrieve_config(adc, &cfg);
		if (err) {
			fprintf(stderr,
				"Cannot retrieve the external trigger configuration (): (%d) %s\n",
				errno, adc_strerror(errno));
		} else {
			err = adc_get_conf(&cfg, ADC_CONF_TRG_THR_ENABLE, &value);
			if (!err)
				fprintf(stdout, "\t\tEnable: %u\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_THR_POLARITY, &value);
			if (!err)
				fprintf(stdout, "\t\tPolarity: %u\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_THR_DELAY, &value);
			if (!err)
				fprintf(stdout, "\t\tDelay: %u\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_THR_THRESHOLD, &value);
			if (!err)
				fprintf(stdout, "\t\tThreshold: %u\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_THR_HYSTERESIS, &value);
			if (!err)
				fprintf(stdout, "\t\tHysteresis: %u\n", value);
		}
	}

	fprintf(stdout, "Trigger (Time) Configuration\n");
	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_TRG_TIM;
	adc_set_conf_mask_all(&cfg, adc);
	err = adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_TRG_THR, &n);
	if (err)
		n = 0;
	for (i = 0; i < n; ++i) {
		fprintf(stdout, "\tTimer trigger %d\n", i + 1);
		cfg.route_to = i;
		err = adc_retrieve_config(adc, &cfg);
		if (err) {
			fprintf(stderr,
				"Cannot retrieve the timer trigger configuration: (%d) %s\n",
			errno, adc_strerror(errno));
		} else {
			err = adc_get_conf(&cfg, ADC_CONF_TRG_TIM_ENABLE, &value);
			if (!err)
				fprintf(stdout, "\t\tEnable: %u\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_TIM_SECONDS, &value);
			if (!err)
				fprintf(stdout, "\t\tSeconds: %u\n", value);
			err = adc_get_conf(&cfg, ADC_CONF_TRG_TIM_NANO_SECONDS, &value);
			if (!err)
				fprintf(stdout, "\t\tNanoseconds: %u\n", value);
		}
	}
}


/**
 * It writes data on file
 * @param[in] filename file where write data
 * @param[in] ch the channel data to write
 * @param[in] pdata data from ADC (is interleaved)
 * @param[in] n_sample number of sample in pdata (is interleaved)
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int write_file(const char *filename, int ch, int16_t *pdata,
		      unsigned int n_sample)
{
	FILE *filp;
	int i;

	int mask = umask(0); /* set umask to have a file with 0666 mode */
	filp = fopen(filename, "w");
	umask(mask); /* restore previous umask */
	if (filp == NULL) {
		printf("fopen %s failed: %s", filename, strerror(errno));
		return -1;
	}
	for (i = 0, pdata += (ch - 1); i < n_sample; ++i, pdata += 4) {
		double val = (*pdata)*bit_scale;
		fprintf(filp, "%d %g\n", i, val);
	}
	if (fclose(filp) < 0)
		return -1;
	return 0;
}


/**
 * It configures the acquisition according to the parameters' string
 * @param[in] adc The ADC device token
 * @param[in] param a parameter string in the form
 *     "<pre-sample>,<post-sample>,<n-shots>,<undersample>"
 * @return 0 on success, otherwise -1 and errno is appropriately set
 *
 * The parameters are not mandatory. If not provided, the application will
 * not try to set them (there is not a default value)
 */
static int fald_acq_acquisition_configuration(struct adc_dev *adc, char *param)
{
	struct adc_conf cfg;
	uint32_t pre, post, nshot, under;
	int ret;

	ret = sscanf(param, "%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32,
		     &pre, &post, &nshot, &under);
	if (ret < 0) {
		errno = EINVAL;
		return -1;
	}

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_ACQ;

	switch (ret) {
	case 4:
		adc_set_conf(&cfg, ADC_CONF_ACQ_UNDERSAMPLE, under);
	case 3:
		adc_set_conf(&cfg, ADC_CONF_ACQ_N_SHOTS, nshot);
	case 2:
		adc_set_conf(&cfg, ADC_CONF_ACQ_POST_SAMP, post);
	case 1:
		adc_set_conf(&cfg, ADC_CONF_ACQ_PRE_SAMP, pre);
		break;
	default:
		errno = EINVAL;
		return -1;
	}

	return adc_apply_config(adc, 0 , &cfg);
}


/**
 * It configures a channel according to the parameters' string
 * @param[in] adc The ADC device token
 * @param[in] param a parameter string in the form
 *     "<channel>,<termination>,<range>,<offset>,<saturation>"
 * @return 0 on success, otherwise -1 and errno is appropriately set
 *
 * The parameters are not mandatory. If not provided, the application will
 * not try to set them (there is not a default value)
 */
static int fald_acq_channel_configuration(struct adc_dev *adc, char *param)
{
	struct adc_conf cfg;
	uint32_t ch, term, range, offset, sat;
	int ret;

	ret = sscanf(param, "%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32,
		     &ch, &term, &range, &offset, &sat);
	if (ret <= 0) { /* channel is mandatory */
		errno = EINVAL;
		return -1;
	}

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_CHN;
	cfg.route_to = ch;

	switch (ret) {
	case 5:
		adc_set_conf(&cfg, ADC_CONF_CHN_SATURATION, sat);
	case 4:
		adc_set_conf(&cfg, ADC_CONF_CHN_OFFSET, offset);
	case 3:
		/*
		 * in-range
		 * 0x23 (35): 100mV range
		 * 0x11 (17): 1V range
		 * 0x45 (69): 10V range
		 * 0x00 (0): Open input
		 */
		switch (range) {
		case 100:
			range = ADC_CONF_100M14B4CHA_CHN_RANGE_100mV;
			bit_scale = 0.05/(1<<15);
			break;
		case 1:
			range = ADC_CONF_100M14B4CHA_CHN_RANGE_1V;
			bit_scale = 0.5/(1<<15);
			break;
		case 10:
			range = ADC_CONF_100M14B4CHA_CHN_RANGE_10V;
			bit_scale = 5.0/(1<<15);
			break;
		}
		adc_set_conf(&cfg, ADC_CONF_CHN_RANGE, range);
	case 2:
		adc_set_conf(&cfg, ADC_CONF_CHN_TERMINATION, term);
		break;
	default:
		errno = EINVAL;
		return -1;
	}

	return adc_apply_config(adc, 0 , &cfg);
}

/**
 * It configures a software trigger according to the parameters' string
 * @param[in] adc The ADC device token
 * @param[in] param a parameter string in the form
 *     "<delay>"
 * @return 0 on success, otherwise -1 and errno is appropriately set
 *
 * The parameters are not mandatory. If not provided, the application will
 * not try to set them (there is not a default value)
 */
static int fald_trg_software_configuration(struct adc_dev *adc, char *param)
{
	uint32_t delay;
	int ret;

	ret = sscanf(param, "%"SCNu32, &delay);
	if (ret <= 0) { /* channel is mandatory */
		errno = EINVAL;
		return -1;
	}

	if (!adc_has_trigger_fire(adc)) {
		fprintf(stderr, "Software trigger not supported\n");
		errno = ADC_ENOP;
		return -1;
	}
	arg_trgsw = 1;

	switch (ret) {
	case 1:
		arg_trgsw_delay = delay;
		break;
	default:
		errno = EINVAL;
		return -1;
	}

	return 0;
}


/**
 * It configures a external trigger according to the parameters' string
 * @param[in] adc The ADC device token
 * @param[in] param a parameter string in the form
 *     "<idx>,<enable>,<polarity>,<delay>"
 * @return 0 on success, otherwise -1 and errno is appropriately set
 *
 * The parameters are not mandatory. If not provided, the application will
 * not try to set them (there is not a default value)
 */
static int fald_trg_external_configuration(struct adc_dev *adc, char *param)
{
	struct adc_conf cfg;
	uint32_t n, enable, polarity, delay;
	int ret;

	ret = sscanf(param, "%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32,
		     &n, &enable, &polarity, &delay);
	if (ret <= 0) { /* channel is mandatory */
		errno = EINVAL;
		return -1;
	}

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_TRG_EXT;
	cfg.route_to = n;

	switch (ret) {
	case 4:
		adc_set_conf(&cfg, ADC_CONF_TRG_EXT_DELAY, delay);
	case 3:
		adc_set_conf(&cfg, ADC_CONF_TRG_EXT_POLARITY, polarity);
	case 2:
		adc_set_conf(&cfg, ADC_CONF_TRG_EXT_ENABLE, enable);
		break;
	default:
		errno = EINVAL;
		return -1;
	}

	return adc_apply_config(adc, 0 , &cfg);
}


/**
 * It configures a threshold trigger according to the parameters' string
 * @param[in] adc The ADC device token
 * @param[in] param a parameter string in the form
 *     "<idx>,<enable>,<polarity>,<threshold>,<hysteresis>,<delay>"
 * @return 0 on success, otherwise -1 and errno is appropriately set
 *
 * The parameters are not mandatory. If not provided, the application will
 * not try to set them (there is not a default value)
 */
static int fald_trg_threshold_configuration(struct adc_dev *adc, char *param)
{
	struct adc_conf cfg;
	uint32_t n, enable, polarity, delay, threshold, hysteresis;
	int ret;

	ret = sscanf(param, "%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32,
		     &n, &enable, &polarity, &threshold, &hysteresis, &delay);
	if (ret <= 0) { /* channel is mandatory */
		errno = EINVAL;
		return -1;
	}

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_TRG_THR;
	cfg.route_to = n;

	switch (ret) {
	case 6:
		adc_set_conf(&cfg, ADC_CONF_TRG_THR_DELAY, delay);
	case 5:
		adc_set_conf(&cfg, ADC_CONF_TRG_THR_HYSTERESIS, hysteresis);
	case 4:
		adc_set_conf(&cfg, ADC_CONF_TRG_THR_THRESHOLD, threshold);
	case 3:
		adc_set_conf(&cfg, ADC_CONF_TRG_THR_POLARITY, polarity);
	case 2:
		adc_set_conf(&cfg, ADC_CONF_TRG_THR_ENABLE, enable);
		break;
	default:
		errno = EINVAL;
		return -1;
	}

	return adc_apply_config(adc, 0 , &cfg);
}


/**
 * It configures a timer trigger according to the parameters' string
 * @param[in] adc The ADC device token
 * @param[in] param a parameter string in the form
 *     "<channel>,<termination>,<range>,<offset>,<saturation>"
 * @return 0 on success, otherwise -1 and errno is appropriately set
 *
 * The parameters are not mandatory. If not provided, the application will
 * not try to set them (there is not a default value)
 */
static int fald_trg_timer_configuration(struct adc_dev *adc, char *param)
{
	struct adc_conf cfg;
	uint32_t n, enable, seconds, nanoseconds = 0;
	int ret;

	ret = sscanf(param, "%"SCNu32",%"SCNu32",%"SCNu32",%"SCNu32,
		     &n, &enable, &seconds, &nanoseconds);
	if (ret <= 2) {
		errno = EINVAL;
		return -1;
	}

	memset(&cfg, 0, sizeof(struct adc_conf));
	cfg.type = ADC_CONF_TYPE_TRG_TIM;
	cfg.route_to = n;

	switch (ret) {
	case 4:
	case 3:
		adc_set_conf(&cfg, ADC_CONF_TRG_TIM_NANO_SECONDS, nanoseconds);
		adc_set_conf(&cfg, ADC_CONF_TRG_TIM_SECONDS, seconds);
		adc_set_conf(&cfg, ADC_CONF_TRG_TIM_ENABLE, enable);
		break;
	default:
		errno = -EINVAL;
		return -1;
	}

	return adc_apply_config(adc, 0 , &cfg);
}

/**
 * It peforms the auto-clear offset according to the command line options
 * @param[in] adc The ADC device token
 * @param[in] argc number of arguments
 * @param[in] argv arguments
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int adc_acq_offset_auto_clear(struct adc_dev *adc,
				     int argc, char *argv[])
{
	char offclr_mode;
	unsigned long flags;
	int c,  opt_index, ret;

	optind = 1; /* set to 1 to make getopt_long happy */
	/* Parse options */
	while ((c = getopt_long(argc, argv, GETOPT_STRING, options, &opt_index))
	       >= 0 ) {
		switch (c) {
		case FALD_ACQ_OPT_OFF_CLR:
			ret = sscanf(optarg, "%c", &offclr_mode);
			if (ret <= 0) {
				errno = EINVAL;
				return -1;
			}
			flags = 0;
			switch (offclr_mode) {
			case 'm':
				flags |= ADC_OFFSET_AC_F_MANUAL;
				break;
			case 'a':
				flags |= ADC_OFFSET_AC_F_RESTORE;
				break;
			case 'z':
				flags |= ADC_OFFSET_AC_F_ZERO;
				break;
			default:
				fprintf(stderr,
					"%s: invalid offset auto-clear mode '%c' (valid modes: {a: automatic, m: manual})",
					_argv[0], offclr_mode);
				return -1;
			}
			ret = adc_offset_auto_clear(adc, flags);
			if (ret < 0)
				return ret;
			break;
		default:
			break;
		}
	}

	return 0;
}

/**
 * It parses command line arguments and consequential it configures the device
 * @param[in] argc number of arguments
 * @param[in] argv arguments
 */
static int fald_acq_parse_args_and_configure(struct adc_dev *adc, int argc, char *argv[])
{
	int c, opt_index;
	int err;

	optind = 1; /* set to 1 to make getopt_long happy */
	/* Parse options */
	while ((c = getopt_long(argc, argv, GETOPT_STRING, options, &opt_index))
	       >= 0 ) {
		switch (c) {
		case 'a':
			err = fald_acq_acquisition_configuration(adc, optarg);
			if (err) {
				fprintf(stderr, "%s: Cannot configure acquisition: %s\n",
					_argv[0], adc_strerror(errno));
				return -1;
			}
			break;
		case 'c':
			err = fald_acq_channel_configuration(adc, optarg);
			if (err) {
				fprintf(stderr, "%s: Cannot configure channel: %s\n",
					_argv[0], adc_strerror(errno));
				return -1;
			}
			break;
		case FALD_ACQ_OPT_TRG_EXT:
			err = fald_trg_external_configuration(adc, optarg);
			if (err) {
				fprintf(stderr, "%s: Cannot configure external trigger: %s\n",
					_argv[0], adc_strerror(errno));
				return -1;
			}
			break;
		case FALD_ACQ_OPT_TRG_THR:
			err = fald_trg_threshold_configuration(adc, optarg);
			if (err) {
				fprintf(stderr, "%s: Cannot configure threshold trigger: %s\n",
					_argv[0], adc_strerror(errno));
				return -1;
			}
			break;
		case FALD_ACQ_OPT_TRG_TIM:
			err = fald_trg_timer_configuration(adc, optarg);
			if (err) {
				fprintf(stderr, "%s: Cannot configure timer trigger: %s\n",
					_argv[0], adc_strerror(errno));
				return -1;
			}
			break;
		case FALD_ACQ_OPT_TRG_SW:
			err = fald_trg_software_configuration(adc, optarg);
			if (err) {
				fprintf(stderr, "%s: Cannot configure software trigger: %s\n",
					_argv[0], adc_strerror(errno));
				return -1;
			}
			break;
		case 'T':
			timeout = atoi(optarg);
			break;
		case 'B':
			binmode = 1; /* do binary (default is 0) */
			basefile = optarg;
			break;
		case 'M':
			binmode = 2; /* do many binaries */
			basefile = optarg;
			break;
		case 'l':
			loop = atoi(optarg);
			break;
		case 's':
			show_ndata = atoi(optarg);
			break;
		default:
			break;
		}
	}

	err = adc_acq_offset_auto_clear(adc, argc, argv);
	if (err)
		return -1;

	fald_acq_print_config(adc);

	return 0;
}


/**
 * It prints data to stdout
 * @param[in] ctrl
 * @param[in] buf buffer to print
 * @param[in] acq_cfg acquisition configuration associated to the buffer
 */
static void fald_acq_print_data(struct adc_buffer *buf,
				struct adc_conf *acq_brd,
				struct adc_conf *acq_cfg,
				unsigned int n)
{
	int j, ch;
	uint32_t pre, nchan;

	if (n == 0)
		return;

	adc_get_conf(acq_brd, ADC_CONF_BRD_N_CHAN, &nchan);
	adc_get_conf(acq_cfg, ADC_CONF_ACQ_PRE_SAMP, &pre);

	/* Print data */
	for (j = 0; j < buf->nsamples; j++) {
		if ( (n > 0 && j < n) ||
		     (n < 0 && (buf->nsamples - j) <= (-n)) ) {
			printf("%5i     ", j - pre);
			for (ch = 0; ch < nchan; ch++) {
				int32_t sample;

				adc_buffer_get_sample(buf, ch, j, &sample);
				printf("%7i", sample);
			}
			printf("\n");
		}
	}
}

/**
 * Save data to a single file
 * @param[in] buf buffer to store
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int fald_acq_write_single(struct adc_buffer *buf)
{
	char fname[PATH_MAX];
	int16_t *data;  /* FMC-ADC-100M sample size is 14bit, 16bit for ZIO */
	FILE *f;
	int err = 0, n;

	sprintf(fname, "%s", basefile);
	f = fopen(fname, "a");
	if (!f) {
		fprintf(stderr, "%s: %s\n",
			fname, strerror(errno));
		exit(1);
	}
#ifdef FIXME
	ctrl = buf->metadata;
	if (fwrite(ctrl, sizeof(struct zio_control), 1, f) != 1)
		err++;
#endif
	data = buf->data;
	n = fwrite(data, buf->samplesize, buf->nsamples, f);
	if (n != buf->nsamples)
		err++;
	fclose(f);

	if (err) {
		fprintf(stderr, "write(%s): short write\n",
			basefile);

		return -1;
	}

	return 0;
}


/**
 * Sava data to different file for control and data for each shot
 * @param[in] buf buffer to store
 * @param[in] shot_i i-th shot acquisition
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static int fald_acq_write_multiple(struct adc_buffer *buf,
				   unsigned int shot_i)
{
	char fname[PATH_MAX];
	uint16_t *data;
	FILE *f;
	int n;

#ifdef FIXME
	/* FIXME zio control is specific. The library should provide
	   at least the metadata size so that automatic tools can at least
	   read */
	sprintf(fname, "%s.%03u.ctrl", basefile, shot_i);
	f = fopen(fname, "w");
	if (!f) {
		fprintf(stderr, "%s: %s\n",
			fname, strerror(errno));
		return -1;
	}
	ctrl = buf->metadata;
	n = fwrite(ctrl, sizeof(struct zio_control), 1, f)
	fclose(f);
	if (n != 1) {
		fprintf(stderr, "write(%s): short write\n", fname);
		return -1;
	}

#endif
	sprintf(fname, "%s.%03u.data", basefile, shot_i);
	f = fopen(fname, "w");
	if (!f) {
		fprintf(stderr, "%s: %s\n",
			fname, strerror(errno));
		return -1;
	}
	data = buf->data;
	n = fwrite(data, buf->samplesize, buf->nsamples, f);
	fclose(f);

	if (n != buf->nsamples) {
		fprintf(stderr, "write(%s): short write\n", fname);
		return -1;
	}

	return 0;
}


/**
 * Plot data using gnuplot
 * @param[in] buf buffer to plot
 * @param[in] ch channel to show
 * @return 0 on success, otherwise -1 and errno is appropriately set
 */
static void fald_acq_plot_data(struct adc_buffer *buf, unsigned int ch)
{
	int16_t *data = buf->data;
	char fname[PATH_MAX];
	char cmd[PATH_MAX + 256];

	snprintf(fname, sizeof(fname), "/tmp/fmcadc.0x%04x.ch%u.dat", devid, ch);
	if (write_file(fname, ch, data, (buf->nsamples)/4) < 0) {
		printf("Cannot plot data. Write data into file %s failed.\n", fname);
		return;
	}
	snprintf(cmd, sizeof(cmd), "echo \"%s plot '%s' with lines "
		 "\" | gnuplot -persist",
		 (arg_x_display == 1 ? "set term x11; " : "set term dumb; "),
		 fname);
	cmd[sizeof(cmd) - 1] = '\0';
	system(cmd);
}

/**
 * It prints data using json format
 */
static void adc_acq_statistics(struct adc_buffer *buf, unsigned int max_chan)
{
	int i;

	fprintf(stdout, "{\"statistics\": [");
	for (i = 0; i < max_chan; ++i) {
		int err;
		int32_t avg;

		fprintf(stdout, "{ \"chan\": %d", i);

		err = adc_buffer_math_avg(buf, i, &avg);
		if (err) {
			fprintf(stdout, ", \"average\": -1");
		} else {
			fprintf(stdout, ", \"average\": %"PRId32, avg);
		}
		if (i == max_chan - 1)
			fprintf(stdout, "}");
		else
			fprintf(stdout, "}, ");
	}
	fprintf(stdout, "]}\n");
}

/**
 * It process the buffer
 * @param[in] buf the buffer to use
 * @param[in] iteration loop iteration number
 */
static void fald_acq_process_buffer(struct adc_buffer *buf,
				    struct adc_conf *cfg_brd,
				    struct adc_conf *cfg_acq,
				    unsigned int iteration)
{
	struct adc_timestamp ts;
	uint32_t nchan;

	adc_tstamp_buffer(buf, &ts);

	/* print/store data*/
	switch(binmode) {
	default:
		fprintf(stderr,
			"time: {secs:%"PRIu64" ticks:%"PRIu64" bins:%"PRIu64"}\n",
			ts.secs, ts.ticks, ts.bins);

		fald_acq_print_data(buf, cfg_brd, cfg_acq, show_ndata);
		break;
	case 1:
		fald_acq_write_single(buf);
		break;
	case 2:
		fald_acq_write_multiple(buf, iteration);
		break;
	}

	adc_get_conf(cfg_brd, ADC_CONF_BRD_N_CHAN, &nchan);

	if (statistics)
		adc_acq_statistics(buf, nchan);

	if (arg_plot) {
		int w;

		for (w = 0; w < nchan; ++w)
			fald_acq_plot_data(buf, w + 1);
	}
}

/**
 * It parses the command line arguments for the basic parameters
 * @param[in] argc number of arguments
 * @param[in] argv arguments
 * @return It returns 0 on success, otherwise -1 and errno is appropriately set
 *
 * It looks for:
 * - device ID
 * - help
 * - version
 */
static int fald_acq_parse_args_basic(int argc, char *argv[])
{
	int opt_index;
	char c;

	while ((c = getopt_long(argc, argv, GETOPT_STRING,
				options, &opt_index)) >= 0 ) {
		switch (c) {
		case 'D': {
			char *strdevid;
			int len;

			strdevid = strchr(optarg, '@');
			if (!strdevid) {
				fprintf(stderr, "%s: invalid device \"%s\" (<type>@<device-id>)\n",
					_argv[0], optarg);
				exit(1);
			}
			strdevid++; /* +1 to skip '@' */
			sscanf(strdevid, "0x%x", &devid);

			/* -1 count '@' it will be replaced by 0
			 * +1 for \0
			 */
			len = (int)(strdevid - 1 - optarg + 1);
			devname = malloc(len);
			if (!devname) {
				fprintf(stderr, "Cannot allocate memory\n");
				exit(1);
			}
			memset(devname, 0, len);
			strncpy(devname, optarg, len - 1);
			break;
		}
		case 'V':
			print_version(argv[0]);
			exit(EXIT_SUCCESS);
			break;
		case 'h':
			fald_help();
			exit(EXIT_SUCCESS);
			break;
		default:
			break;
		}
	}
	return 0;
}


/**
 * Within this function we do a single acquisition
 * @param[in] adc ADC device token
 * @param[in] buf buffer to use
 * @param[in] cfg_acq acquisition configuration
 * @param[in] iteration acquisition number
 * @return 0 on success
 */
static int adc_acq_acquisition(struct adc_dev *adc,
			       struct adc_buffer *buf,
			       struct adc_conf *cfg_brd,
			       struct adc_conf *cfg_acq,
			       unsigned int iteration)
{
	struct timeval tv = {0, 0};
	uint32_t nshots, pre, post;
	int err, k;

	fprintf(stderr, "Acquisition %u\n", iteration + 1);

	/* Stop any pending acquisition */
	err = adc_acq_stop(adc, 0);
	if (err) {
		fprintf(stderr,
			"Cannot stop pending acquisition: (%d) %s\n",
			errno, adc_strerror(errno));
		goto err;
	}

	err = adc_acq_start(adc, ADC_F_FLUSH, &tv);
	if (err) {
		fprintf(stderr, "Cannot start acquisition: (%d) %s\n",
			errno, adc_strerror(errno));
		goto err;
	}

	adc_get_conf(cfg_acq, ADC_CONF_ACQ_PRE_SAMP, &pre);
	adc_get_conf(cfg_acq, ADC_CONF_ACQ_POST_SAMP, &post);
	adc_get_conf(cfg_acq, ADC_CONF_ACQ_N_SHOTS, &nshots);
	for (k = 0; k < nshots; ++k) {
		fprintf(stderr,
			"shot: %i/%u, nsamples: %u\n",
			k + 1, nshots, pre + post);

		if (arg_trgsw) {
			sleep(arg_trgsw_delay);
			err = adc_trigger_fire(adc);
			if (err)
				goto err;
		}

		tv.tv_sec = 60;
		err = adc_fill_buffer(adc, buf,
				      fixup ? ADC_F_FIXUP: 0,
				      &tv);
		if (err) {
			fprintf(stderr,
				"Failed to retrieve for data: (%d) %s\n",
				errno, adc_strerror(errno));
			goto err;
		}

		fald_acq_process_buffer(buf, cfg_brd, cfg_acq, iteration);
	}
	return 0;

err:
	return err;
}

/**
 * Within this function we repeat the same acquisition a number of times
 * @param[in] adc ADC device token
 * @param[in] buf buffer to use
 * @param[in] cfg_acq acquisition configuration
 * @param[in] n number of acquisition to do
 * @return 0 on success
 */
static int adc_acq_acquisition_n(struct adc_dev *adc,
				 struct adc_conf *cfg_brd,
				 struct adc_conf *cfg_acq,
				 unsigned int n)
{
	struct adc_buffer *buf;
	uint32_t pre, post;
	int i, err = 0;

	adc_get_conf(cfg_acq, ADC_CONF_ACQ_PRE_SAMP, &pre);
	adc_get_conf(cfg_acq, ADC_CONF_ACQ_POST_SAMP, &post);

	buf = adc_request_buffer(adc, pre + post, NULL /* alloc */, 0);
	if (!buf) {
		fprintf(stderr, "Cannot allocate buffer (%s)\n",
			adc_strerror(errno));
		return -1;
	}

	for (i = 0; i < n && arg_no_read == 0; ++i) {
		err = adc_acq_acquisition(adc, buf, cfg_brd, cfg_acq, i);
		if (err)
			break;
	}

	adc_release_buffer(adc, buf, NULL);

	return err;
}


/**
 * It configures an acquisition using user's parameters
 * @param[in] adc ADC device token
 * @param[in] argc number of parameters
 * @param[in] argv user's parameters
 * @param[out] cfg_acq acquisition configuration
 * @return 0 on success
 */
static int adc_acq_configure(struct adc_dev *adc,
			     int argc, char *argv[],
			     struct adc_conf *cfg_brd,
			     struct adc_conf *cfg_acq)
{
	int err = 0;

	/* get the new given trigger and acq config
	   Only the ones provided will override the current ones */
	err = fald_acq_parse_args_and_configure(adc, argc, argv);
	if (err)
		return err;

	/*
	 * Within this program we already know the configuration (we
	 * applied it before), but here we will retrieve it from the
	 * library anyway because the library might have changed something
	 */
	memset(cfg_brd, 0, sizeof(struct adc_conf));
	cfg_brd->type = ADC_CONF_TYPE_BRD;
	adc_set_conf_mask_all(cfg_brd, adc);
	err = adc_retrieve_config(adc, cfg_brd);
	if (err) {
		fprintf(stderr,
			"Cannot retrive board configuration: (%d) %s\n",
			errno, adc_strerror(errno));
		return err;
	}

	memset(cfg_acq, 0, sizeof(struct adc_conf));
	cfg_acq->type = ADC_CONF_TYPE_ACQ;
	adc_set_conf_mask_all(cfg_acq, adc);
	err = adc_retrieve_config(adc, cfg_acq);
	if (err) {
		fprintf(stderr,
			"Cannot retrive acquisition configuration: (%d) %s\n",
			errno, adc_strerror(errno));
		return err;
	}

	return err;
}

/**
 * It cleans up our internal mess on exit
 */
static void adc_acq_cleanup(void)
{
	free(devname);
}

int main(int argc, char *argv[])
{
	struct adc_dev *adc;
	struct adc_conf cfg_acq, cfg_brd;
	int  err;

	atexit(adc_acq_cleanup);

	/* set local _argv[0] with  pg name */
	_argv[0] = argv[0];
	err = fald_acq_parse_args_basic(argc, argv);
	if (err)
		exit(err);

	if (!devname) {
		fprintf(stderr,
			"%s: device is a mandatory argument (%s@0x%x)\n",
			argv[0], devname, devid);
		fald_help();
		goto err_devid_invalid;
	}

	err = adc_init();
	if (err)
		exit(EXIT_FAILURE);

	/* Open the ADC */
	adc = adc_open(devname, devid, 0, 0, ADC_F_FLUSH);
	if (!adc) {
		fprintf(stderr, "%s: cannot open device: %s\n",
			argv[0], adc_strerror(errno));
		goto err_open;
	}
	err = adc_acq_configure(adc, argc, argv, &cfg_brd, &cfg_acq);
	if (err)
		goto err_cfg;
	err = adc_acq_acquisition_n(adc, &cfg_brd, &cfg_acq, loop);
	if (err)
		goto err_acq;
	adc_close(adc);
	adc_exit();

	exit(EXIT_SUCCESS);

err_acq:
err_cfg:
	adc_close(adc);
err_open:
err_devid_invalid:
	adc_exit();
	exit(EXIT_FAILURE);
}
